# -*- coding: utf-8 -*-

"""
/***************************************************************************
 OTBAlgorithmProvider.py
 -----------------------
        date                 : 2018-01-30
        copyright            : (C) 2018 by CNES
        email                : rkm
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
"""

__author__ = 'Rashad Kanavath'
__date__ = '2018-01-30'
__copyright__ = '(C) 2018 by CNES'

# This will get replaced with a git SHA1 when you do a git archive

__revision__ = '$Format:%H$'

import os
import re
from qgis.PyQt.QtGui import QIcon
from qgis.PyQt.QtCore import QCoreApplication
from qgis.core import (Qgis, QgsApplication, QgsProcessingProvider, QgsMessageLog)
from qgis import utils

from processing.core.ProcessingConfig import ProcessingConfig, Setting

from otb import OTBUtils
from otb.OTBAlgorithm import OTBAlgorithm

def otb_exe_file(f):
    if os.name == 'nt':
        return f + '.exe'
    else:
        return f

class OTBAlgorithmProvider(QgsProcessingProvider):
    def __init__(self):
        super().__init__()
        self.algs = []
        #!hack for 6.6!#
        self.version = '6.6.0'
        self.descriptionFile = ''

    def canBeActivated(self):
        if not self.isActive():
            return False
        folder = OTBUtils.otbFolder()
        if folder and os.path.exists(folder):
            if os.path.isfile(self.algsFile(folder)):
                return True
            utils.iface.messageBar().pushWarning("OTB", "Cannot find '{}'. OTB provider will be disabled".format(self.algsFile(folder)))
        self.setActive(False)
        return False

    def check_app_folder(self, v):
        if not self.isActive():
            return
        if not v:
            self.setActive(False)
            raise ValueError(self.tr('Cannot activate OTB provider'))

        folder = OTBUtils.otbFolder()
        otb_app_dirs = self.otb_app_dir_list(v)
        if len(otb_app_dirs) < 1:
            self.setActive(False)
            raise ValueError(self.tr("'{}' does not exist. OTB provider will be disabled".format(v)))

        #isValid is True if there is atleast one valid otb application is given path
        isValid = False
        descr_folder = self.descrFolder(folder)
        for app_dir in otb_app_dirs:
            if not os.path.exists(app_dir):
                continue
            for otb_app in os.listdir(app_dir):
                if not otb_app.startswith('otbapp_') or \
                    'TestApplication' in otb_app or \
                    'ApplicationExample' in otb_app:
                    continue
                app_name = os.path.basename(otb_app).split('.')[0][7:]
                dfile = os.path.join(descr_folder, app_name + '.txt')
                isValid = True
                if not os.path.exists(dfile):
                    cmdlist = [os.path.join(
                        folder, 'bin',
                        otb_exe_file('otbQgisDescriptor')),
                               app_name, app_dir, descr_folder + '/']
                    commands = ' '.join(cmdlist)
                    QgsMessageLog.logMessage(self.tr(commands), self.tr('Processing'), Qgis.Critical)
                    OTBUtils.executeOtb(commands, feedback=None)

        if isValid:
            utils.iface.messageBar().pushInfo("OTB", "OTB provider is activated from '{}'.".format(folder))
        else:
            self.setActive(False)
            raise ValueError(self.tr("No OTB algorithms found in '{}'. OTB will be disabled".format(','.join(otb_app_dirs))))

    def normalize_path(self, p):
        # https://stackoverflow.com/a/20713238/1003090
        return os.path.normpath(os.sep.join(re.split(r'\\|/', p)))

    def check_otb_folder(self, v):
        if not self.isActive():
            return
        if not v or not os.path.exists(v):
            self.setActive(False)
            raise ValueError(self.tr("'{}' does not exist. OTB provider will be disabled".format(v)))
        path = self.normalize_path(v)
        if not os.path.exists(os.path.join(path, 'bin', otb_exe_file('otbApplicationLauncherCommandLine'))):
            self.setActive(False)
            raise ValueError(self.tr("Cannot find '{}'. OTB will be disabled".format(os.path.join(v, 'bin', otb_exe_file('otbApplicationLauncherCommandLine')))))

    def write_otbcli_wrapper_script(self):
        otb_folder = self.normalize_path(OTBUtils.otbFolder())
        otb_app_path_env = os.pathsep.join(self.otb_app_dir_list(OTBUtils.appFolder()))
        gdal_data_dir = None
        geotiff_csv_dir = None
        try:
            if os.name == 'nt':
                app_vargs = " %*"
                export_cmd = 'SET '
                first_line = ':: Setup environment for OTB package. Generated by QGIS plugin'
                otb_app_launcher = os.path.join(otb_folder, 'bin', 'otbApplicationLauncherCommandLine.exe')
                gdal_data_dir = os.path.join(otb_folder, 'share', 'data')
                geotiff_csv_dir = os.path.join(otb_folder, 'share', 'epsg_csv')
            else:
                app_vargs = " $@"
                export_cmd = 'export '
                first_line = '#!/bin/sh'
                otb_app_launcher = os.path.join(otb_folder, 'bin', 'otbApplicationLauncherCommandLine')
                lines = None
                env_profile = os.path.join(otb_folder, 'otbenv.profile')
                with open(env_profile) as f:
                    lines = f.readlines()
                    lines = [x.strip() for x in lines]
                    for line in lines:
                        if not line or line.startswith('#'):
                            continue
                        if 'GDAL_DATA=' in line:
                            gdal_data_dir = line.split("GDAL_DATA=")[1]
                        if 'GEOTIFF_CSV='in line:
                            geotiff_csv_dir = line.split("GEOTIFF_CSV=")[1]

            with open(OTBUtils.OTB_CLI_FILE, 'w') as otb_cli_file:
                otb_cli_file.write(first_line + os.linesep)
                otb_cli_file.write(export_cmd + "LC_NUMERIC=C" + os.linesep)
                otb_cli_file.write(export_cmd + "GDAL_DRIVER_PATH=disable" + os.linesep)
                otb_cli_file.write(export_cmd + "\"GDAL_DATA=" + gdal_data_dir + "\"" + os.linesep)
                otb_cli_file.write(export_cmd + "\"GEOTIFF_CSV=" + geotiff_csv_dir + "\"" + os.linesep)
                otb_cli_file.write(export_cmd + "\"OTB_APPLICATION_PATH=" + otb_app_path_env + "\"" + os.linesep)
                otb_cli_file.write("\""  + otb_app_launcher + "\"" + app_vargs  + os.linesep)

            if not os.name == 'nt':
                os.chmod(OTBUtils.OTB_CLI_FILE, 0o744)
        except BaseException as e:
            import traceback
            os.remove(OTBUtils.OTB_CLI_FILE)
            errmsg = "Cannot write:" + OTBUtils.OTB_CLI_FILE + "\nError:\n" +  traceback.format_exc()
            QgsMessageLog.logMessage(self.tr(errmsg), self.tr('Processing'), Qgis.Critical)
            raise e

    def load(self):
        group = self.name()
        ProcessingConfig.settingIcons[group] = self.icon()
        ProcessingConfig.addSetting(Setting(group, OTBUtils.ACTIVATE, self.tr('Activate'), True))
        ProcessingConfig.addSetting(Setting(group, OTBUtils.FOLDER,
                                            self.tr("OTB folder"),
                                            OTBUtils.otbFolder(),
                                            valuetype=Setting.FOLDER,
                                            validator=self.check_otb_folder
                                           ))
        ProcessingConfig.addSetting(Setting(group, OTBUtils.APP_FOLDER,
                                            self.tr("OTB application folder"),
                                            OTBUtils.appFolder(),
                                            valuetype=Setting.MULTIPLE_FOLDERS,
                                            validator=self.check_app_folder
                                           ))
        ProcessingConfig.addSetting(Setting(group, OTBUtils.SRTM_FOLDER,
                                            self.tr("SRTM tiles folder"),
                                            OTBUtils.srtmFolder(),
                                            valuetype=Setting.FOLDER
                                           ))
        ProcessingConfig.addSetting(Setting(group, OTBUtils.GEOID_FILE,
                                            self.tr("Geoid file"),
                                            OTBUtils.geoidFile(),
                                            valuetype=Setting.FOLDER
                                           ))

        ProcessingConfig.readSettings()
        self.refreshAlgorithms()
        return True

    def unload(self):
        for setting in [OTBUtils.ACTIVATE, OTBUtils.FOLDER, OTBUtils.SRTM_FOLDER, OTBUtils.GEOID_FILE]:
            ProcessingConfig.removeSetting(setting)

    def isActive(self):
        return ProcessingConfig.getSetting(OTBUtils.ACTIVATE)

    def setActive(self, active):
        ProcessingConfig.setSettingValue(OTBUtils.ACTIVATE, active)

    def createAlgsList(self):
        algs = []
        try:
            folder = OTBUtils.otbFolder()
            alg_names = []
            algs_txt = self.algsFile(folder)
            with open(algs_txt) as lines:
                line = lines.readline().strip('\n').strip()
                if line != '' and line.startswith('#'):
                    line = lines.readline().strip('\n').strip()
                while line != '' and not line.startswith('#'):
                    data = line.split('|')
                    self.descriptionFile = self.descrFile(folder, str(data[1]) + '.txt')
                    group, name = str(data[0]), str(data[1])
                    if name not in alg_names:
                        algs.append(OTBAlgorithm(group, name, self.descriptionFile))
                        #avoid duplicate algorithms from algs.txt file (possible but rare)
                        alg_names.append(name)
                    line = lines.readline().strip('\n').strip()
        except Exception as e:
            import traceback
            errmsg = "Could not open OTB algorithm from file: \n" + self.descriptionFile + "\nError:\n" +  traceback.format_exc()
            QgsMessageLog.logMessage(self.tr(errmsg), self.tr('Processing'), Qgis.Critical)
        return algs

    def loadAlgorithms(self):
        if not self.canBeActivated():
            return

        version_file = os.path.join(OTBUtils.otbFolder(), 'share', 'doc', 'otb', 'VERSION')
        if not os.path.isfile(version_file):
            version_file = os.path.join(OTBUtils.otbFolder(), 'VERSION')

        if os.path.isfile(version_file):
            with open(version_file) as vf:
                vlines = vf.readlines()
                vlines = [l.strip() for l in vlines]
                vline = vlines[0]
                if 'OTB Version:' in vline:
                    self.version = vline.split(':')[1].strip()

        QgsMessageLog.logMessage(self.tr("Loading OTB '{}'.".format(self.version)), self.tr('Processing'), Qgis.Info)
        self.algs = self.createAlgsList()
        for a in self.algs:
            self.addAlgorithm(a)
        self.algs = []

        folder = OTBUtils.otbFolder()
        cli_ext = ''
        env_profile = os.path.join(folder, 'otbenv.profile')
        if os.name == 'nt':
            cli_ext = '.bat'
            env_profile = os.path.join(folder, 'otbenv.bat')

        if os.path.exists(env_profile) and os.path.isfile(env_profile):
            OTBUtils.OTB_CLI_FILE = os.path.normpath(
                os.path.join(
                    QgsApplication.qgisSettingsDirPath(),
                    'processing', 'qgis_otb_cli' + cli_ext)
                    )
            #if not os.path.exists(OTBUtils.OTB_CLI_FILE):
            #rewrite always for now..
            self.write_otbcli_wrapper_script()
        else:
            OTBUtils.OTB_CLI_FILE = os.path.join(folder, 'bin', 'otbcli'  + cli_ext)

        QgsMessageLog.logMessage(self.tr("Using  otbcli: '{}'.".format(OTBUtils.OTB_CLI_FILE)), self.tr('Processing'), Qgis.Info)

    def algsFile(self, d):
        return os.path.join(self.descrFolder(d), 'algs.txt')

    def descrFolder(self, d):
        #!hack for 6.6!#
        if os.path.exists(os.path.join(d, 'description')):
            return os.path.join(d, 'description')
        else:
            return os.path.join(d, 'share', 'otb', 'description')

    def descrFile(self, d, f):
        return os.path.join(self.descrFolder(d), f)

    def otb_app_dir_list(self, v):
        #!hack needed for QGIS < 3.2!#
        v = v.replace(';', os.pathsep)
        #!hack needed for QGIS < 3.2!#
        folders = v.split(os.pathsep)
        app_dirs = []
        for f in folders:
            if f is not None and os.path.exists(f):
                app_dirs.append(self.normalize_path(f))
        return app_dirs

    def name(self):
        return 'OTB'

    def longName(self):
        return 'OTB ({})'.format(self.version) if self.version is not None else 'OTB'

    def id(self):
        return 'otb'

    def supportsNonFileBasedOutput(self):
        """
        OTB Provider doesn't support non file based outputs
        """
        return False

    def icon(self):
        pluginPath = os.path.split(os.path.dirname(__file__))[0]
        return QIcon(os.path.join(pluginPath, 'otb', 'otb.png'))

    def tr(self, string, context=''):
        if context == '':
            context = 'OTBAlgorithmProvider'
        return QCoreApplication.translate(context, string)

    def defaultVectorFileExtension(self, hasGeometry=True):
        return 'shp'

    def defaultRasterFileExtension(self):
        return 'tif'

    def supportedOutputTableExtensions(self):
        return ['dbf']
